open Client

(* A few manually constructed terms. *)

let hole =
  ML.Hole []

let x =
  ML.Var "x"

let y =
  ML.Var "y"

let id =
  ML.Abs ("x", x)

let delta =
  ML.Abs ("x", ML.App (x, x))

let deltadelta =
  ML.App (delta, delta)

let idid =
  ML.App (id, id)

let k =
  ML.Abs ("x", ML.Abs ("y", x))

let genid =
  ML.Let ("x", id, x)

let genidid =
  ML.Let ("x", id, ML.App (x, x))

let genkidid =
  ML.Let ("x", ML.App (k, id), ML.App (x, id))

let genkidid2 =
  ML.Let ("x", ML.App (ML.App (k, id), id), x)

let app_pair = (* ill-typed *)
  ML.App (ML.Tuple [id; id], id)

let unit =
  ML.Tuple []

let abs_match_with =
  ML.(
    Abs(
      "x",
      Match(
        Tuple[],
        [ (PTuple[], Tuple[]) ]
      )
    )
  )

(* option *)
let option_env =

  (* type 'a option = None | Some of 'a *)
  let option_typedecl =
    let open Datatype in {
    name = Type "option";
    type_params = [ "'a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label"None";
        type_name = Type "option";
        arg_type = None;
      } ; {
        label_name = Label "Some";
        type_name = Type "option";
        arg_type = Some (ML.TyVar "'a");
      }
    ]
  } in
  Datatype.Env.add_decl Datatype.Env.empty option_typedecl

let none = ML.Variant ( Datatype.Label "None" , None )

let some =
  ML.Variant (
    Datatype.Label "Some",
    Some id
  )

let match_none = ML.(
    Match (none, [
      PVariant (Datatype.Label "None", None), none ;
      PVariant (Datatype.Label "Some", Some (PVar "x")), x ;
    ])
  )

let match_some = ML.(
    Match (some, [
      PVariant (Datatype.Label "None", None), none ;
      PVariant (Datatype.Label "Some", Some PWildcard), none
    ])
  )

let match_some_annotated = ML.(
    Match (some, [
      ( PVariant (Datatype.Label "None", None), none );
      ( PAnnot (PVariant (Datatype.Label "Some", Some PWildcard),
               (["'a"], TyConstr (Datatype.Type "option", [ TyVar "'a" ]))),
        none );
    ])
  )

(* list *)
let list_env =

  (* type 'a list = Nil | Cons of 'a * 'a list *)
  let list_typedecl =
    let open Datatype in {
    name = Type "list";
    type_params = [ "'a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label "Nil";
        type_name = Type "list";
        arg_type = None;
      } ; {
        label_name = Label "Cons";
        type_name = Type "list";
        arg_type = Some (ML.(TyProduct [ TyVar "'a" ; TyConstr (Type "list", [ TyVar "'a" ]) ]));
      }
    ]
  } in

  Datatype.Env.add_decl option_env list_typedecl

let nil = ML.Variant ( Datatype.Label "Nil" , None )

let list =
  ML.Variant (
    Datatype.Label "Cons",
    Some (ML.Tuple [ id ; nil ])
  )

let list_annotated =
  let open ML in
  Annot (
    Variant (
      Datatype.Label "Cons",
      Some (Tuple [ Annot (id, (["'a"], TyArrow (TyVar "'a", TyVar "'a")));
                    nil ])
    ),
    (["'a"; "'b"], TyConstr (Datatype.Type "list", [TyArrow (TyVar "'a", TyVar "'b")]))
  )

(* tree *)
let tree_env =

  (* type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree *)
  let tree_typedecl =
    let open Datatype in {
    name = Type "tree";
    type_params = [ "'a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label "Leaf";
        type_name = Type "tree";
        arg_type = None
      } ; {
        label_name = Label "Node";
        type_name = Type "tree";
        arg_type = Some (ML.(TyProduct [
          TyConstr (Type "tree", [ TyVar "'a" ]);
          TyVar "'a";
          TyConstr (Type "tree", [ TyVar "'a" ]);
        ]))
      }
    ];
  } in

  Datatype.Env.add_decl list_env tree_typedecl

let leaf = ML.Variant ( Datatype.Label "Leaf" , None )

let node =
  ML.Variant (
    Datatype.Label "Node",
    Some (ML.Tuple [
      leaf ;
      id ;
      leaf ;
    ])
  )

let test =
  Test.(Log.with_log CheckML.test_with_log)

let () =
  assert (test Datatype.Env.empty hole);
  assert (test Datatype.Env.empty id);
  assert (test Datatype.Env.empty idid);
  assert (test Datatype.Env.empty genid);
  assert (test Datatype.Env.empty genidid);
  assert (test Datatype.Env.empty genkidid);
  assert (test Datatype.Env.empty genkidid2);
  assert (test option_env none);
  assert (test option_env some);
  assert (test list_env nil);
  assert (test list_env list);
  assert (test tree_env leaf);
  assert (test tree_env node);
  assert (test Datatype.Env.empty abs_match_with);
  assert (test option_env match_none);
  assert (test option_env match_some);
  (* we include some printing below because Dune
     appears to only show tests that show some output. *)
  print_endline "TestML: all tests passed."


(*************************************************************************)

exception ParsingError

(* Main parsing function *)
let parse = MLParser.prog MLLexer.read

let ast_from_string s =
  let lexbuf = Lexing.from_string s in
  try parse lexbuf
  with _ -> raise ParsingError

let letify (typedecl_env, xts) =
  let rec aux xts =
    match xts with
    | [] ->
        assert false
    | [(_, t)] ->
        t
    | (x, t) :: xts ->
        ML.Let (x, t, aux xts)
  in
  (typedecl_env, aux xts)

let test_ok ?(typedecl="") s expected =
  let s = typedecl ^ "\nlet _ = " ^ s in
  let (datatype_env, t) = ast_from_string s |> letify in
  OUnit2.assert_equal ~printer:MLPrinter.to_string expected t;
  OUnit2.assert_bool "typing" (test datatype_env t)

let test_error_parsing ?(typedecl="") (s : string) =
  let s = typedecl ^ "\nlet _ = " ^ s in
  OUnit2.assert_raises ParsingError (fun () -> ast_from_string s)

let test_id _ =
  test_ok "fun x -> x" id

let test_delta_delta_error _ =
  test_error_parsing "(fun x -> x x (fun x -> x x)"

let test_idid _ =
  test_ok "(fun x -> x) (fun x -> x)" idid

let test_idid_error _ =
  test_error_parsing "fun x -> x fun x -> x"

let test_unit _ =
  test_ok "()" unit

let test_abs_match_with _ =
  test_ok "fun x -> match () with () -> () end" abs_match_with

let test_let _ =
  test_ok "let y = fun x -> x in ()" (ML.Let("y", id, unit))

let test_let_prod_singleton _ =
  test_ok "let (y,) = (fun x -> x,) in ()"
    (ML.LetProd (["y"], ML.Tuple [id], unit))

let test_let_prod _ =
  test_ok "let (y,z) = (fun x -> x, ()) in ()"
  (ML.LetProd (["y";"z"], ML.Tuple [id;unit], unit))

let test_singleton _ =
  test_ok
    "(fun x -> x,)"
    (ML.Tuple [id])

let test_pair_tuple _ =
  test_ok
    "(fun x -> x, fun x -> x)"
    (ML.Tuple [id; id])

let option_env_str =
  "type option 'a = None | Some of 'a"

let test_none _ =
  test_ok
    ~typedecl:option_env_str
    "None"
    none

let test_some _ =
  test_ok
    ~typedecl:option_env_str
    "Some (fun x -> x)"
    some

let test_some_pair _ =
  test_ok
    ~typedecl:option_env_str
    "Some (fun x -> x, fun x -> x)"
    (ML.Variant (Datatype.Label "Some", Some (ML.Tuple [id;id])))

let list_env_str = "type list 'a = Nil | Cons of {'a * list 'a}"

let test_list_nil _ =
  test_ok
    ~typedecl:list_env_str
    "Nil"
    nil

let test_list_cons _ =
  test_ok
    ~typedecl:list_env_str
    "Cons (fun x -> x, Nil)"
    list

let test_arrow _ =
  test_ok
    ~typedecl:"type func 'a 'b = Func of 'a -> 'b"
    "Func (fun x -> x)"
    (ML.Variant (Datatype.Label "Func", Some id))

let test_match_tuple _ =
  test_ok
    "match (fun x -> x, ()) with (f, ()) -> f end"
    (ML.Match
       (ML.Tuple [id;unit],
        [ (ML.PTuple[ML.PVar "f"; ML.PTuple[]], ML.Var "f") ]
    ))

let test_match_none _ =
  test_ok
    ~typedecl:option_env_str
    "match None with
     | None -> None
     | Some x -> x
     end"
  match_none

let test_match_some _ =
  test_ok
    ~typedecl:option_env_str
    "match Some (fun x -> x) with
     | None -> None
     | Some _ -> None
     end"
  match_some

let test_match_some_annotated _ =
  test_ok
    ~typedecl:option_env_str
    "match Some (fun x -> x) with
     | None -> None
     | (Some _ : some 'a. option 'a) -> None
     end"
  match_some_annotated

let test_suite = OUnit2.(
  "ML test suite" >::: [
      "id" >:: test_id;
      "delta delta error" >:: test_delta_delta_error;
      "id id" >:: test_idid;
      "id id error" >:: test_idid_error;
      "unit" >:: test_unit;
      "abs match with" >:: test_abs_match_with;
      "let" >:: test_let;
      "let prod singleton" >:: test_let_prod_singleton;
      "let prod" >:: test_let_prod;
      "singleton" >:: test_singleton;
      "pair tuple" >:: test_pair_tuple;
      "none" >:: test_none;
      "some" >:: test_some;
      "some pair" >:: test_some_pair;
      "list nil" >:: test_list_nil;
      "list cons" >:: test_list_cons;
      "arrow" >:: test_arrow;
      "match tuple" >:: test_match_tuple;
      "match none" >:: test_match_none;
      "match some" >:: test_match_some;
      "match some annotated" >:: test_match_some_annotated;
    ]
  )

let () =
  OUnit2.run_test_tt_main test_suite
